@@ -32,10 +32,6 @@ namespace embree
       throw std::runtime_error("unknown scene format: " + filename.ext());
   }
 
-  void tab(std::ostream& cout, int n) {
-    for (int i=0; i<n; i++) cout << "  ";
-  }
-
   void SceneGraph::PerspectiveCameraNode::print(std::ostream& cout, int depth) {
     cout << "PerspectiveCameraNode @ " << this << " { " << std::endl;
     if (name != "") {
@@ -68,6 +64,18 @@ namespace embree
     tab(cout, depth); cout << "}" << std::endl;
   }
 
+  void SceneGraph::MultiTransformNode::print(std::ostream& cout, int depth)
+  {
+    cout << "MultiTransformNode @ " << this << " { " << std::endl;
+    tab(cout, depth+1); cout << "closed = " << closed << std::endl;
+    tab(cout, depth+1); cout << "numInstances = " << spaces.size() << std::endl;
+    if (spaces.size() > 0) {
+      tab(cout, depth+1); cout << "numTimeSteps = " << spaces[0].size() << std::endl;
+    }
+    tab(cout, depth+1); cout << "child = "; child->print(cout,depth+1);
+    tab(cout, depth); cout << "}" << std::endl;
+  }
+
   void SceneGraph::GroupNode::print(std::ostream& cout, int depth)
   {
     cout << "GroupNode @ " << this << " { " << std::endl;
@@ -220,6 +228,23 @@ namespace embree
     }
   }
 
+  void SceneGraph::MultiTransformNode::calculateStatistics(Statistics& stat)
+  {
+    indegree++;
+    if (indegree == 1)
+    {
+      child->calculateStatistics(stat);
+
+      for (size_t i = 0; i < spaces.size(); ++i) {
+        stat.numTransformNodes++;
+        if (child->indegree == 1)
+          stat.numTransformedObjects++;
+
+        if (spaces[i].size() > 1) child->calculateStatistics(stat); // break instance up when motion blur is used
+      }
+    }
+  }
+
   void SceneGraph::GroupNode::calculateStatistics(Statistics& stat)
   {
     indegree++;
@@ -313,6 +338,16 @@ namespace embree
     }
   }
 
+  void SceneGraph::MultiTransformNode::calculateInDegree()
+  {
+    indegree++;
+    if (indegree == 1) {
+      child->calculateInDegree();
+      for (size_t i = 0; i < spaces.size(); ++i)
+        if (spaces[i].size() > 1) child->calculateInDegree(); // break instance up when motion blur is used
+    }
+  }
+
   void SceneGraph::GroupNode::calculateInDegree()
   {
     indegree++;
@@ -365,13 +400,25 @@ namespace embree
     return closed && (indegree == 1);
   }
 
+  bool SceneGraph::MultiTransformNode::calculateClosed(bool group_instancing)
+  {
+    assert(indegree);
+    if (!closed) {
+      closed = group_instancing;
+      closed &= child->calculateClosed(group_instancing);
+      hasLightOrCamera = child->hasLightOrCamera;
+    }
+    return closed && (indegree == 1);
+
+  }
+
   bool SceneGraph::GroupNode::calculateClosed(bool group_instancing)
   {
     assert(indegree);
     if (!closed) {
       closed = group_instancing;
       hasLightOrCamera = false;
-      for (auto c : children) {
+      for (auto& c : children) {
         closed &= c->calculateClosed(group_instancing);
         hasLightOrCamera |= c->hasLightOrCamera;
       }
@@ -444,6 +491,16 @@ namespace embree
     indegree--;
   }
 
+  void SceneGraph::MultiTransformNode::resetInDegree()
+  {
+    closed = false;
+    if (indegree == 1) {
+      child->resetInDegree();
+      if (spaces.size() > 1) child->resetInDegree(); // break instance up when motion blur is used
+    }
+    indegree--;
+  }
+
   void SceneGraph::GroupNode::resetInDegree()
   {
     closed = false;
@@ -466,7 +523,7 @@ namespace embree
       if (n.size() && n.size() != N)
         THROW_RUNTIME_ERROR("incompatible vertex array sizes");
     if (texcoords.size() && texcoords.size() != N) THROW_RUNTIME_ERROR("incompatible vertex array sizes");
-    for (auto tri : triangles) {
+    for (auto& tri : triangles) {
       if (size_t(tri.v0) >= N || size_t(tri.v1) >= N || size_t(tri.v2) >= N)
         THROW_RUNTIME_ERROR("invalid triangle");
     }
@@ -484,7 +541,7 @@ namespace embree
       if (n.size() && n.size() != N)
         THROW_RUNTIME_ERROR("incompatible vertex array sizes");
     if (texcoords.size() && texcoords.size() != N) THROW_RUNTIME_ERROR("incompatible vertex array sizes");
-    for (auto quad : quads) {
+    for (auto& quad : quads) {
       if (size_t(quad.v0) >= N || size_t(quad.v1) >= N || size_t(quad.v2) >= N || size_t(quad.v3) >= N)
         THROW_RUNTIME_ERROR("invalid quad");
     }
@@ -496,7 +553,7 @@ namespace embree
     for (const auto& p : positions) 
       if (p.size() != N) 
         THROW_RUNTIME_ERROR("incompatible vertex array sizes");
-    for (auto grid : grids) {
+    for (auto& grid : grids) {
       if (size_t(grid.startVtx) >= N || size_t(grid.lineStride) >= N || size_t(grid.resX) >= 0x7fff || size_t(grid.resY) >= 0x7fff)
         THROW_RUNTIME_ERROR("invalid grid");
     }
@@ -519,7 +576,7 @@ namespace embree
       if (size_t(i) >= texcoords.size()) THROW_RUNTIME_ERROR("invalid texcoord index array");
     for (auto i : holes) 
       if (size_t(i) >= verticesPerFace.size()) THROW_RUNTIME_ERROR("invalid hole index array");
-    for (auto crease : edge_creases) 
+    for (auto& crease : edge_creases) 
       if (max(size_t(crease.x),size_t(crease.y)) >= N) THROW_RUNTIME_ERROR("invalid edge crease array");
     if (edge_crease_weights.size() != edge_creases.size())
       THROW_RUNTIME_ERROR("invalid edge crease weight array");
@@ -593,13 +650,13 @@ namespace embree
         type == RTC_GEOMETRY_TYPE_ROUND_HERMITE_CURVE ||
         type == RTC_GEOMETRY_TYPE_NORMAL_ORIENTED_HERMITE_CURVE)
     {
-      for (auto hair : hairs)
+      for (auto& hair : hairs)
         if (size_t(hair.vertex+1) >= N)
           THROW_RUNTIME_ERROR("invalid hair");
     }
     else 
     {
-      for (auto hair : hairs)
+      for (auto& hair : hairs)
         if (size_t(hair.vertex+3) >= N)
           THROW_RUNTIME_ERROR("invalid hair");
     }
@@ -967,35 +1024,35 @@ namespace embree
     else if (Ref<SceneGraph::TriangleMeshNode> mesh = node.dynamicCast<SceneGraph::TriangleMeshNode>()) 
     {
       avector<Vec3fa> positions1;
-      for (auto P : mesh->positions.back()) 
+      for (auto& P : mesh->positions.back()) 
         positions1.push_back(P+dP);
       mesh->positions.push_back(std::move(positions1));
     }
     else if (Ref<SceneGraph::QuadMeshNode> mesh = node.dynamicCast<SceneGraph::QuadMeshNode>()) 
     {
       avector<Vec3fa> positions1;
-      for (auto P : mesh->positions.back()) 
+      for (auto& P : mesh->positions.back()) 
         positions1.push_back(P+dP);
       mesh->positions.push_back(std::move(positions1));
     }
     else if (Ref<SceneGraph::GridMeshNode> mesh = node.dynamicCast<SceneGraph::GridMeshNode>()) 
     {
       avector<Vec3fa> positions1;
-      for (auto P : mesh->positions.back()) 
+      for (auto& P : mesh->positions.back()) 
         positions1.push_back(P+dP);
       mesh->positions.push_back(std::move(positions1));
     }
     else if (Ref<SceneGraph::HairSetNode> mesh = node.dynamicCast<SceneGraph::HairSetNode>()) 
     {
       avector<Vec3ff> positions1;
-      for (auto P : mesh->positions.back()) 
+      for (auto& P : mesh->positions.back()) 
         positions1.push_back(P+Vec3ff(dP,0.0f));
       mesh->positions.push_back(std::move(positions1));
     }
     else if (Ref<SceneGraph::PointSetNode> mesh = node.dynamicCast<SceneGraph::PointSetNode>())
     {
       avector<Vec3ff> positions1;
-      for (auto P : mesh->positions.back())
+      for (auto& P : mesh->positions.back())
         positions1.push_back(P+Vec3ff(dP,0.0f));
       mesh->positions.push_back(std::move(positions1));
 
@@ -1005,7 +1062,7 @@ namespace embree
     else if (Ref<SceneGraph::SubdivMeshNode> mesh = node.dynamicCast<SceneGraph::SubdivMeshNode>())
     {
       avector<Vec3fa> positions1;
-      for (auto P : mesh->positions.back())
+      for (auto& P : mesh->positions.back())
         positions1.push_back(P+dP);
       mesh->positions.push_back(std::move(positions1));
     }
@@ -1141,6 +1198,10 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->spaces.time_range = time_range;
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      for (size_t i = 0; i < xfmNode->spaces.size(); ++i)
+        xfmNode->spaces[i].time_range = time_range;
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1233,6 +1294,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->child = convert_triangles_to_quads(xfmNode->child,prop);
     } 
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      xfmNode->child = convert_triangles_to_quads(xfmNode->child,prop);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1283,6 +1347,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->child = convert_quads_to_grids(xfmNode->child, resX, resY);
     } 
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      xfmNode->child = convert_quads_to_grids(xfmNode->child, resX, resY);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1327,7 +1394,10 @@ namespace embree
   {
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->child = convert_grids_to_quads(xfmNode->child);
-    } 
+    }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      xfmNode->child = convert_grids_to_quads(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1452,6 +1522,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->child = my_merge_quads_to_grids(xfmNode->child);
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      xfmNode->child = my_merge_quads_to_grids(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1532,7 +1605,10 @@ namespace embree
   {
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->child = convert_quads_to_subdivs(xfmNode->child);
-    } 
+    }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      xfmNode->child = convert_quads_to_subdivs(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1574,6 +1650,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->child = convert_bezier_to_lines(xfmNode->child);
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      xfmNode->child = convert_bezier_to_lines(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1586,7 +1665,7 @@ namespace embree
       for (auto& p : hmesh->positions)
         lmesh->positions.push_back(p);
 
-      for (auto hair : hmesh->hairs) {
+      for (auto& hair : hmesh->hairs) {
         lmesh->hairs.push_back(SceneGraph::HairSetNode::Hair(hair.vertex+0,hair.id));
         lmesh->hairs.push_back(SceneGraph::HairSetNode::Hair(hair.vertex+1,hair.id));
         lmesh->hairs.push_back(SceneGraph::HairSetNode::Hair(hair.vertex+2,hair.id));
@@ -1601,6 +1680,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->child = convert_flat_to_round_curves(xfmNode->child);
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      xfmNode->child = convert_flat_to_round_curves(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1625,6 +1707,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       xfmNode->child = convert_round_to_flat_curves(xfmNode->child);
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      xfmNode->child = convert_round_to_flat_curves(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1649,6 +1734,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       convert_bezier_to_bspline(xfmNode->child);
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      convert_bezier_to_bspline(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1666,6 +1754,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       convert_bspline_to_bezier(xfmNode->child);
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      convert_bspline_to_bezier(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1682,6 +1773,9 @@ namespace embree
     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
       convert_bezier_to_hermite(xfmNode->child);
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      convert_bezier_to_hermite(xfmNode->child);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
       for (size_t i=0; i<groupNode->children.size(); i++) 
@@ -1706,6 +1800,16 @@ namespace embree
       }
       xfmNode->child = remove_mblur(xfmNode->child, mblur);
     }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      if (mblur) {
+        if (xfmNode->spaces.size() == 0 || xfmNode->spaces[0].size() > 1)
+          return nullptr;
+      } else {
+        if (xfmNode->spaces.size() > 0 && xfmNode->spaces[0].size() > 1)
+          return node;
+      }
+      xfmNode->child = remove_mblur(xfmNode->child, mblur);
+    }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>())
     {
       for (size_t i=0; i<groupNode->children.size(); i++)
@@ -1741,9 +1845,14 @@ namespace embree
 
   void SceneGraph::convert_mblur_to_nonmblur(Ref<Node> node)
   {
-     if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
-       xfmNode->spaces.spaces.resize(1);
-       convert_mblur_to_nonmblur(xfmNode->child);
+    if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
+      xfmNode->spaces.spaces.resize(1);
+      convert_mblur_to_nonmblur(xfmNode->child);
+    }
+    else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+      for (size_t i = 0; i < xfmNode->spaces.size(); ++i)
+        xfmNode->spaces[i].spaces.resize(1);
+      convert_mblur_to_nonmblur(xfmNode->child);
     }
     else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) 
     {
@@ -1871,7 +1980,11 @@ namespace embree
     {
       if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
         convertGeometries(group,xfmNode->child, spaces*xfmNode->spaces);
-      } 
+      }
+      else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+        for (size_t i = 0; i < xfmNode->spaces.size(); ++i)
+          convertGeometries(group,xfmNode->child, spaces*xfmNode->spaces[i]);
+      }
       else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) {
         for (const auto& child : groupNode->children) convertGeometries(group,child,spaces);
       }
@@ -1907,6 +2020,24 @@ namespace embree
       return object_mapping[node];
     }
 
+    void convertInstances(std::vector<Ref<SceneGraph::Node>>& group, const Ref<SceneGraph::Node>& node, const std::vector<SceneGraph::Transformations>& spaces)
+    {
+      if (node->isClosed()) {
+        //if (group.size() % 10000 == 0) std::cout << "." << std::flush;
+        group.push_back(new SceneGraph::MultiTransformNode(spaces,lookupGeometries(node)));
+      }
+      else if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
+        for (size_t i = 0; i < spaces.size(); ++i)
+          convertInstances(group,xfmNode->child, spaces[i]*xfmNode->spaces);
+      }
+      else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+        convertInstances(group,xfmNode->child, spaces*xfmNode->spaces);
+      }
+      else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) {
+        for (const auto& child : groupNode->children) convertInstances(group,child,spaces);
+      }
+    }
+
     void convertInstances(std::vector<Ref<SceneGraph::Node>>& group, const Ref<SceneGraph::Node>& node, const SceneGraph::Transformations& spaces)
     {
       if (node->isClosed()) {
@@ -1915,7 +2046,10 @@ namespace embree
       }
       else if (Ref<SceneGraph::TransformNode> xfmNode = node.dynamicCast<SceneGraph::TransformNode>()) {
         convertInstances(group,xfmNode->child, spaces*xfmNode->spaces);
-      } 
+      }
+      else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>()) {
+        convertInstances(group,xfmNode->child, spaces*xfmNode->spaces);
+      }
       else if (Ref<SceneGraph::GroupNode> groupNode = node.dynamicCast<SceneGraph::GroupNode>()) {
         for (const auto& child : groupNode->children) convertInstances(group,child,spaces);
       }
@@ -1952,7 +2086,13 @@ namespace embree
         auto new_node = new SceneGraph::TransformNode(xfmNode->spaces,convertMultiLevelInstances(xfmNode->child));
         object_mapping[node] = new_node;
         group.push_back(new_node);
-      } 
+      }
+      else if (Ref<SceneGraph::MultiTransformNode> xfmNode = node.dynamicCast<SceneGraph::MultiTransformNode>())
+      {
+        auto new_node = new SceneGraph::MultiTransformNode(xfmNode->spaces,convertMultiLevelInstances(xfmNode->child));
+        object_mapping[node] = new_node;
+        group.push_back(new_node);
+      }
     }
       
     Ref<SceneGraph::Node> convertMultiLevelInstances(const Ref<SceneGraph::Node>& node)
